from langchain.prompts import PromptTemplate
import dataclasses
import pydantic
from enum import Enum
from loguru import logger
from typing import Dict,Any,List,Optional,Union,Tuple,Callable,Annotated
from byzerllm.utils import (function_calling_format,
                            response_class_format,
                            response_class_format_after_chat                            
                            )

class Role:
    User = "user"
    Assistant = "assistant"
    System = "system"

@dataclasses.dataclass
class LLMHistoryItem:
      role: str
      content: str

@dataclasses.dataclass
class LLMResponse:
    output: Union[str,List[float]]
    input: Union[str,Dict[str,Any]]
    metadata: Dict[str,Any] = dataclasses.field(default_factory=dict)


class LLMFunctionCallResponse(pydantic.BaseModel):
    response:LLMResponse
    values:List[Any]
    metadata:Dict[str,Any]


class LLMClassResponse(pydantic.BaseModel):
    response:LLMResponse
    value:Optional[Any]
    metadata:Dict[str,Any]

@dataclasses.dataclass
class LLMRequest:
    instruction: Union[str,List[str]]
    embedding: bool = False
    max_length: int = 4096
    top_p: float = 0.95
    temperature: float = 0.1    
        

@dataclasses.dataclass
class FintuneRequestExtra:
    max_seq_length: int = 1024
    num_train_epochs: int = 1
    logging_steps: int = 100
    save_steps: int = 100
    extra_params: Dict[str,Any] = dataclasses.field(default_factory=dict)

@dataclasses.dataclass
class  FintuneRequest:
    model_path: str
    pretrained_model_type: str
    input_data_path: str
    extra_params: FintuneRequestExtra = dataclasses.field(default_factory=FintuneRequestExtra)


class InferBackend:
    Transformers = "transformers"
    VLLM = "ray/vllm"
    LLAMA_CPP = "llama_cpp"
    DeepSpeed = "ray/deepspeed"

@dataclasses.dataclass
class ExecuteCodeResponse:
      status: int
      output: str      
      code: str
      prompt: str
      variables: Dict[str,Any]=dataclasses.field(default_factory=dict)

class EventName(Enum):
    BEFORE_CALL_MODEL = "before_call_model"
    AFTER_CALL_MODEL = "after_call_model"

EventCallbackResult = Tuple[bool, Optional[Any]]
EventCallback = Callable[..., EventCallbackResult]    

class Template:
    def __init__(self,
                 role_mapping:Dict[str,str],
                 generation_config:Dict[str,Any],
                 clean_func:Callable[[str],str]=lambda s: s,
                 function_calling_format_func=function_calling_format,
                 response_class_format_func=response_class_format,
                 response_class_format_after_chat_func=response_class_format_after_chat
                 ) -> None:
        self.role_mapping = role_mapping
        self.generation_config = generation_config
        self.clean_func = clean_func        
        self.function_calling_format_func = function_calling_format_func
        self.response_class_format_func = response_class_format_func
        self.response_class_format_after_chat_func = response_class_format_after_chat_func


class Templates:

    def default_format(t,v):
        return f"{t}{v}"


    @staticmethod
    def qwen():
        def clean_func(v):            
            if "<|im_end|>" in v:
                v = v.split("<|im_end|>")[0]
            if "<|endoftext|>" in v:
                v = v.split("<|endoftext|>")[0] 
            if "<|im_start|>" in v:             
                v = v.split("<|im_start|>")[0]   
            return v   

        def sys_format(t,v):
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)


        return Template(role_mapping={
                        "user_role":"<|im_start|>user\n",
                        "assistant_role": "<|im_end|>\n<|im_start|>assistant\n",
                        "system_msg":"<|im_start|>system\n{system_msg}<|im_end|>",
                        "system_msg_func":sys_format
                        },
                        generation_config={                            
                            "generation.repetition_penalty":1.1,
                            "generation.stop_token_ids":[151643,151645]},                  
                        clean_func=clean_func) 
    
    @staticmethod
    def llama():
        def sys_format(t,v):
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)
        
        def user_format(t,v):
            return f"<s>[INST] {v} [/INST]"
        
        def assistant_format(t,v):
            return f" {v} </s>"
        
        return Template(
            role_mapping={
               "user_role":"",
               "assistant_role": "",
               "system_msg":"<s>[INST] <<SYS>>\n{system_msg}\n<</SYS>>\n[/INST]</s>",
               "system_msg_func":sys_format,
               "user_role_func": user_format,
               "assistant_role_func": assistant_format
            },            
            generation_config={},
            clean_func=lambda s: s
        )
    
    @staticmethod
    def deepseek_code_chat():
        '''
        DeepSeek Coder Chat mode template:

        ### Instruction:
        ['content']
        ### Response:
        ['content']
        <|EOT|>
        ### Instruction:
        ['content']
        ### Response:
        '''
        

        def sys_format(t:Annotated[str,"the field system_msg in role_mapping "],
                       v:Annotated[str,"the system message in chat"]):
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)
        
        def user_format(t:Annotated[str,"the field user_role in role_mapping"],
                        v:Annotated[str,"the user message in chat"]):
            '''
            format single user message
            '''
            return f"### Instruction:\n{v}"
        
        def assistant_format(t:Annotated[str,"the field assistant_role in role_mapping"],
                             v:Annotated[str,"the assistant message in chat"]):
            '''
            format single assitant message.
            
            Notice that here we do not use `t` , because we will
            use the `t` as the final suffix.
            '''
            return f"### Response:\n{v}\n<|EOT|>"
        
        return Template(
            role_mapping={
               "user_role":"",
               "assistant_role": "### Response:\n",
               "system_msg":"{system_msg}",
               "system_msg_func":sys_format,
               "user_role_func": user_format,
               "assistant_role_func": assistant_format
            },            
            generation_config={"generation.stop_token_ids":[32021]},
            clean_func=lambda s: s
        )
    @staticmethod
    def deepseek_code_insertion():        
        def sys_format(t,v):
            if "<｜fim▁hole｜>" not in v:
                raise Exception("the system message should contains <｜fim▁hole｜>")
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)
        
        def user_format(t,v):            
            return ""
        
        def assistant_format(t,v):            
            return ""
        
        return Template(
            role_mapping={
               "user_role":"",
               "assistant_role": "",
               "system_msg":"<｜fim▁begin｜>{system_msg}<｜fim▁end｜>",
               "system_msg_func":sys_format,
               "user_role_func": user_format,
               "assistant_role_func": assistant_format
            },            
            generation_config={},
            clean_func=lambda s: s
        )
    
    @staticmethod
    def deepseek_code_completion():        
        def sys_format(t,v):            
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)
        
        def user_format(t,v):            
            return ""
        
        def assistant_format(t,v):            
            return ""
        
        return Template(
            role_mapping={
               "user_role":"",
               "assistant_role": "",
               "system_msg":"{system_msg}",
               "system_msg_func":sys_format,
               "user_role_func": user_format,
               "assistant_role_func": assistant_format
            },            
            generation_config={},
            clean_func=lambda s: s
        )
    @staticmethod
    def yi():
        def clean_func(v):                    
            return v   

        def sys_format(t,v):
            m = PromptTemplate.from_template(t)
            return m.format(system_msg=v)


        return Template(role_mapping={
                        "user_role":"<|im_start|>user\n",
                        "assistant_role": "<|im_end|>\n<|im_start|>assistant\n",
                        "system_msg":"<|im_start|>system\n{system_msg}<|im_end|>",
                        "system_msg_func":sys_format
                        },
                        generation_config={"generation.stop_token_ids":[7]},                  
                        clean_func=clean_func) 

    @staticmethod
    def default():
        def clean_func(v):                    
            return v   

        def sys_format(t,v):
            return v

        return Template(role_mapping={
                        "user_role":"User:",
                        "assistant_role": "Assistant:",
                        "system_msg":"You are a helpful assistant. Think it over and answer the user question correctly.",
                        "system_msg_func":sys_format
                        },
                        generation_config={},                  
                        clean_func=clean_func)   

    @staticmethod
    def empty():
        def clean_func(v):                    
            return v   

        def sys_format(t,v):
            return v

        return Template(role_mapping={
                        "user_role":"",
                        "assistant_role": "",
                        "system_msg":"",
                        "system_msg_func":sys_format
                        },
                        generation_config={},                  
                        clean_func=clean_func)

class MessageRole(str, Enum):
    """Message role."""

    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    FUNCTION = "function"
    TOOL = "tool"
    CHATBOT = "chatbot"  

DEFAULT_CONTEXT_WINDOW = 8192  # tokens
DEFAULT_NUM_OUTPUTS = 256  # tokens        

class LLMMetadata(pydantic.BaseModel):
    context_window: int = pydantic.Field(
        default=DEFAULT_CONTEXT_WINDOW,
        description=(
            "Total number of tokens the model can be input and output for one response."
        ),
    )
    num_output: int = pydantic.Field(
        default=DEFAULT_NUM_OUTPUTS,
        description="Number of tokens the model can output when generating a response.",
    )
    is_chat_model: bool = pydantic.Field(
        default=False,
        description=(
            "Set True if the model exposes a chat interface (i.e. can be passed a"
            " sequence of messages, rather than text), like OpenAI's"
            " /v1/chat/completions endpoint."
        ),
    )
    is_function_calling_model: bool = pydantic.Field(
        default=False,
        # SEE: https://openai.com/blog/function-calling-and-other-api-updates
        description=(
            "Set True if the model supports function calling messages, similar to"
            " OpenAI's function calling API. For example, converting 'Email Anya to"
            " see if she wants to get coffee next Friday' to a function call like"
            " `send_email(to: string, body: string)`."
        ),
    )
    model_name: str = pydantic.Field(
        default="unknown",
        description=(
            "The model's name used for logging, testing, and sanity checking. For some"
            " models this can be automatically discerned. For other models, like"
            " locally loaded models, this must be manually specified."
        ),
    )
    system_role: MessageRole = pydantic.Field(
        default=MessageRole.SYSTEM,
        description="The role this specific LLM provider"
        "expects for system prompt. E.g. 'SYSTEM' for OpenAI, 'CHATBOT' for Cohere",
    )  
    class Config:
        protected_namespaces = ()   


import json
import os
from typing import List, Dict
from os.path import expanduser

from typing import List, Optional
from pydantic import BaseModel

class CollectionItem(BaseModel):
    name: str
    description: Optional[str] = ""    


class CollectionManager:
   def __init__(self,base_dir:str):
        home = expanduser("~")        
        base_dir = base_dir or os.path.join(home, ".auto-coder")
        self.collections_json_file = os.path.join(base_dir, "storage", "data", "collections.json") 
        self.collections_json_file_dir = os.path.join(base_dir, "storage", "data")
        
        if not os.path.exists(self.collections_json_file_dir):
            os.makedirs(self.collections_json_file_dir)

        self.collections: Dict[str, CollectionItem] = {}
        self.load_collections()

   def load_collections(self):
        if os.path.exists(self.collections_json_file):
            with open(self.collections_json_file, "r") as f:
                collections_data = json.load(f)
                for name, item_data in collections_data.items():
                    self.collections[name] = CollectionItem(**item_data)
        if "default" not in self.collections:
            self.collections["default"] = CollectionItem(name="default", description="Default collection")            
                  

   def save_collections(self):
       collections_data = {name: item.model_dump() for name, item in self.collections.items()}
       with open(self.collections_json_file, "w") as f:
           json.dump(collections_data, f, indent=2,ensure_ascii=False)

   def add_collection(self, collection_item: CollectionItem):
       self.collections[collection_item.name] = collection_item
       self.save_collections()

   def update_collection(self, collection_item: CollectionItem):
       if collection_item.name in self.collections:
           self.collections[collection_item.name] = collection_item
           self.save_collections()
       else:
           raise ValueError(f"Collection '{collection_item.name}' not found.")

   def delete_collection(self, name: str):
       if name in self.collections:
           del self.collections[name]
           self.save_collections()
       else:
           raise ValueError(f"Collection '{name}' not found.")

   def get_collection(self, name: str) -> CollectionItem:
       if name in self.collections:
           return self.collections[name]
       else:
           raise ValueError(f"Collection '{name}' not found.")

   def check_collection_exists(self, name: str) -> bool:
       return name in self.collections   

   def get_all_collections(self) -> List[CollectionItem]:
       return list(self.collections.values())


from pydantic import BaseModel
from byzerllm.utils.nontext import Image,Audio


class Bool(BaseModel):
    value: bool


class Int(BaseModel):
    value: int


class Float(BaseModel):
    value: float


class ImagePath(BaseModel):
    value: str

    def __str__(self):
        return Image.load_image_from_path(self.value)


class AudioPath(BaseModel):
    value: str

    def __str__(self):
        return Audio.load_audio_from_path(self.value)



from typing import List, Optional, Tuple,Any,Dict,Callable,Generator
from transformers import AutoTokenizer, AutoModelForCausalLM,BitsAndBytesConfig
import ray
import torch
import deepspeed
import deepspeed.comm as dist
import sentencepiece as spm
import numpy as np
import datetime
import uuid
import json
import os
from pyjava.storage import streaming_tar as STar
from pyjava import RayContext
from pyjava.api.mlsql import DataServer
from byzerllm import BlockRow
from ray.air.util.torch_dist import (
    ActorHandle,
    _get_node_and_gpu_ids,
    _init_torch_distributed,
    get_address_and_port,
)
import dataclasses
from ... import print_flush
import shutil
from ray.train.constants import DEFAULT_NCCL_SOCKET_IFNAME

DEFUALT_CONFIG = '''
{
  "gradient_accumulation_steps": 1,
  "train_micro_batch_size_per_gpu": 1,
  "prescale_gradients": false,
  "zero_allow_untested_optimizer": true,
  "optimizer": {
    "type": "AdamW",
    "params": {
      "lr": 1e-8,
      "eps": 1.0e-8,
      "betas": [
        0.9,
        0.95
      ],
      "weight_decay": 0.1
    }
  },
  "tensorboard": {
    "enabled": true    
  },
  "zero_optimization": {
    "stage": 3,
    "offload_optimizer": {
         "device": "cpu"         
     },           
    "offload_param": {
         "device": "cpu"
    },
    "contiguous_gradients": true,
    "allgather_bucket_size": 1e8,
    "reduce_bucket_size": 1e8,
    "overlap_comm": true,
    "reduce_scatter": true
  },
  "steps_per_print": 16,
  "gradient_clipping": 1.0,
  "wall_clock_breakdown": true,
  "bf16": {
    "enabled": true
  }
}
'''

@dataclasses.dataclass
class TrainArgs:
    model_path: str = "" 
    tokenizer_path: str = ""
    sft_name: str = ""
    steps_per_epoch: int = 4096
    is_partition_data: bool = False
    epoches:int = 1
    checkpoint_saving_path: str = "/home/byzerllm/data/checkpoints"
    max_length: int = 4096
    data_dir: str = "/home/byzerllm/data/raw_data"
    data_mode: str = "auto"
     

@dataclasses.dataclass
class DeviceID:
    node_id: int
    gpu_ids: List[int]
    rank: int

class DataEngine():
    def __init__(self, data_dir, tokenizer_path, micro_batch_size, max_length,world_size,rank):
        self.MIN_TEXT_LEN = 20
        self.EOS_TOKEN_ID = 2
        self.data_dir = data_dir
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(tokenizer_path)
        self.micro_batch_size = micro_batch_size
        self.max_length = max_length
        self.data = []
        self.global_input_paths = [self.data_dir + "/" + x
                                   for x in os.listdir(self.data_dir)]
        self.local_input_paths = [x for i, x in
                                  enumerate(self.global_input_paths)
                                  if i % world_size == rank]

    def load_data(self):
        for file_path in self.local_input_paths:
            data = []
            with open(file_path, encoding="utf-8", errors="ignore") as f:
                for line_id, line in enumerate(f):
                    cc = self.sp.EncodeAsIds(line.strip()) + [self.EOS_TOKEN_ID]
                    if len(cc) < self.MIN_TEXT_LEN:
                        cc = []
                    data.extend(cc)
                    if len(data) >= self.micro_batch_size * (self.max_length + 1):
                        index = self.micro_batch_size * (self.max_length + 1)
                        self.data.append(data[:index])
                        data = []
        return
    
    def reset(self):
        self.data = []
        self.load_data()

    def get_data(self):
        data = self.data.pop(0)
        seq = np.asarray(data).reshape(self.micro_batch_size, self.max_length + 1)
        data = torch.LongTensor(seq)
        data = data.cuda(non_blocking=True)
        return data

class ParallelConfig:
    """Configuration for the distributed execution.    
    """

    def __init__(
        self,
        num_workers:int,            
        get_model:Callable[[str,Dict],Any],        
        ds_config:Dict[Any,Any],         
        data_refs:List[DataServer] = [],
        train_args = TrainArgs(),            
        backend: str = "nccl",  
        setup_nccl_socket_ifname_by_ip:bool = False
    ) -> None:
        self.world_size = num_workers        
        self.backend = backend
        self.ds_config = ds_config if ds_config else json.loads(DEFUALT_CONFIG)
        self.train_args = train_args  
        self.get_model = get_model 
        self.data_refs = data_refs 
        # if the nodes in cluster  have different network interface name, we need to set the NCCL_SOCKET_IFNAME 
        # manually otherwise you may meet the following error in deepspeed:
        # torch.distributed.DistBackendError: NCCL error in: ../torch/csrc/distributed/c10d/ProcessGroupNCCL.cpp:1275, 
        # internal error, NCCL version 2.14.3
        # ncclInternalError: Internal check failed.
        # Last error:
        # Proxy Call to rank 8 failed (Connect)
        self.setup_nccl_socket_ifname_by_ip = setup_nccl_socket_ifname_by_ip    
    
def _init_distributed_environment(
        parallel_config: ParallelConfig,
        rank: int,
        distributed_init_method: str        
    ) -> None:        
        if parallel_config.backend == "nccl":
            # Same as in Ray Train
            os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "1"
            # All workers on a same node should share the same set of
            # visible GPUs. Otherwise they can't talk among themselves.
            # os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str(gid) for gid in gpu_ids)
            if "NCCL_SOCKET_IFNAME" not in os.environ:
                os.environ["NCCL_SOCKET_IFNAME"] = DEFAULT_NCCL_SOCKET_IFNAME

        # os.environ["RANK"] = str(rank)
        # os.environ["LOCAL_RANK"] = str(rank)
        # os.environ["WORLD_SIZE"] = str(parallel_config.world_size)
        # os.environ["LOCAL_WORLD_SIZE"] = str(parallel_config.world_size)        
        print(f'''
deepspeed worker config:
              RANK:{rank} 
              WORLD_SIZE:{parallel_config.world_size}
              CUDA_VISIBLE_DEVICES:{os.environ["CUDA_VISIBLE_DEVICES"]} 
              LOCAL_RANK:{os.environ["LOCAL_RANK"]}
              LOCAL_WORLD_SIZE:{os.environ["LOCAL_WORLD_SIZE"]}
              NCCL_SOCKET_IFNAME：{os.environ["NCCL_SOCKET_IFNAME"]}
''',flush=True) 
        # torch.cuda.set_device(rank)
        """Initialize the distributed environment."""
        deepspeed.init_distributed(
            dist_backend="nccl",
            auto_mpi_discovery=False,
            verbose=True,
            init_method=distributed_init_method,
            rank=rank,
            world_size=parallel_config.world_size,
        )
        # torch.distributed.init_process_group(
        #     backend="nccl",
        #     world_size=parallel_config.world_size,
        #     rank=rank,
        #     init_method=distributed_init_method,            
        # )
        # # A small all_reduce for warmup.
        # torch.distributed.all_reduce(torch.zeros(1).cuda())

        

class ResourceWorker:
    def __init__(
        self,        
        parallel_config: ParallelConfig,        
        rank: int               
    ) -> None:
        self.parallel_config = parallel_config        
        self.rank = rank        
        self.ds_config = self.parallel_config.ds_config

    def get_node_and_gpu_ids(self):
        """Returns the node and GPU ids of the current worker."""
        node_id, gpu_ids = _get_node_and_gpu_ids()
        return DeviceID(node_id, gpu_ids, self.rank)  

    def rank(self):
        return self.rank  
    
    def get_node_ip_address(self):
        return ray.util.get_node_ip_address()
    
    def get_address_and_port(self):
        return get_address_and_port()
    
    def get_network_interface(self):
        import netifaces
        interfaces = netifaces.interfaces()
        target_iface = ""
        for iface in interfaces:
            addrs = netifaces.ifaddresses(iface)
            if netifaces.AF_INET in addrs:
                ip = addrs[netifaces.AF_INET][0]['addr']
                address = self.get_node_ip_address()
                if ip == address:
                    target_iface = iface
                    break
        return target_iface        


class Worker:
    
    def __init__(
        self,        
        parallel_config: ParallelConfig,        
        rank: int,
        distributed_init_method:str                
       
    ) -> None:
        self.parallel_config = parallel_config        
        self.rank = rank        
        self.ds_config = self.parallel_config.ds_config
        self.get_model = self.parallel_config.get_model
        self.distributed_init_method = distributed_init_method 
        self.data_dir = os.path.join(self.parallel_config.train_args.data_dir,f"data-{self.rank}") 
        
        # if the data is not from data_refs(from Byzer) , it may
        # means that the data is prepared in every node before run the training.
        # we just respect the data_dir provied by the user.                
        if not self.parallel_config.data_refs:                  
            self.data_dir = self.parallel_config.train_args.data_dir

        self.model = None
        self.tokenizer = None
        self.tensorboard_pid = None        
    
    def get_node_and_gpu_ids(self):
        """Returns the node and GPU ids of the current worker."""
        node_id, gpu_ids = _get_node_and_gpu_ids()
        return DeviceID(node_id, gpu_ids, self.rank)        
    
    def setup_tensorboard(self)->Optional[Tuple[str,int]]:
        #         "tensorboard": {
        #     "enabled": true,
        #     "output_path": "/home/byzerllm/data/train_ck/logs/",
        #     "job_name": "7b-pt"
        # },
        tensorboard_config = self.ds_config.get("tensorboard", {"enabled":False})
        if tensorboard_config["enabled"]:   
            import subprocess               
            ip, port = get_address_and_port()
            log_dir = tensorboard_config["output_path"]
            job_name = tensorboard_config["job_name"]
            log_dir = os.path.join(log_dir,job_name)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            tb_process = subprocess.Popen(['tensorboard', '--logdir', log_dir,"--port",str(port),"--host",ip], stdout=subprocess.PIPE, stderr=subprocess.PIPE)                                    
            self.tensorboard_pid = tb_process.pid
            return (ip,port)
        return None

    def _train(self,data_engine, model_engine):
        model_engine.train()
        step = 0
        while step < self.parallel_config.train_args.steps_per_epoch:
            data = data_engine.get_data()
            loss = model_engine(data, labels=data).loss
            model_engine.backward(loss)
            model_engine.step()
            step += 1
        return  

    def get_checkpoint(self):
        if self.rank == 0:
            sft_name = self.parallel_config.train_args.sft_name
            final_path = self.parallel_config.train_args.checkpoint_saving_path
            # get the last checkpoint
            # the checkpoint path is like this:
            # /home/byzerllm/data/sft-20230805-1224-30-173a8dca-9e4a-411c-9fcb-fc979e3460f6/finetune_model/Epoch-1
            # /home/byzerllm/data/sft-20230805-1224-30-173a8dca-9e4a-411c-9fcb-fc979e3460f6/finetune_model/Epoch-2
            # get the last one
            dirs = os.listdir(final_path)
            dirs.sort(key=lambda x: int(x.split("-")[-1]))
            final_path = os.path.join(final_path,dirs[-1])

            result = []
            count = 0
            print_flush(f"[{sft_name}] Store model({final_path}) to Ray object store")
            for item in STar.build_rows_from_file(final_path):
                if count % 1000 == 0:
                    print_flush(f"[{sft_name}] Progress: {count} processed")
                count += 1    
                result.append(ray.put(item))
            
            print_flush(f'''
                [{sft_name}] Train Actor already finished.
                [{sft_name}] It may take a while to transfer the model from Ray object store to delta lake. 
                [{sft_name}] Try to check the progress in Byzer console or Byzer Notebook. 
                ''')    
            return (result,count) 
        return None


    
    def train(self):        
        data_engine = self.prepare_data()
        model_engine = self.prepare_model()        
        epoch = 0
        while epoch < self.parallel_config.train_args.epoches:
            self._train(data_engine, model_engine)
            epoch += 1
            model_engine.save_checkpoint(f"{self.parallel_config.train_args.checkpoint_saving_path}",
                                        tag=f"Epoch-{epoch}")
            data_engine.reset()
            
    def prepare_data(self):        
        
        if self.parallel_config.data_refs: 
            if not os.path.exists(self.data_dir):
                os.makedirs(self.data_dir)

            train_file = os.path.join(self.data_dir,f"train.txt")
            
            '''
            simplely write data to text file
            may need to be think how to handle for new line if the conversation contains new line. This is because
            the data_engine will get data line by line util touch the limit of max_length, then this seuqence will be
            used to train the model.
            But since this is for pretraining, it should be fine.
            '''
            with open(train_file,"w") as f: 
                count = 0
                data_ref = self.parallel_config.data_refs[self.rank]
                print(f"Start to read data to {data_ref.host}:{data_ref.port}. target file:{train_file}",flush=True)
                for item in RayContext.collect_from([data_ref]):                
                    if "conversation" in item:
                        item["conversation"] = item["conversation"].tolist()
                        s =  " ".join(conversation)
                        f.write(s+"\n")                    
                    elif "history" in item:
                        # support alpaca format data
                        conversation = [sub.tolist() for sub in item["history"].tolist()]
                        conversation = [{"human":x[0],"assistant":x[1]} for x in conversation]
                        latest_conversation = [{"human":item["instruction"],"assistant":item["output"]}] if "instruction" in item and item["instruction"] else []
                        s = " ".join(conversation) + " ".join(latest_conversation)
                        f.write(s+"\n")
                    elif "text" in item:
                        f.write(item["text"]+"\n")
                    else:
                        raise Exception("Unknow data format")                             
                    count += 1         

        
        tokenizer_path = self.parallel_config.train_args.tokenizer_path        
        micro_batch_size = self.ds_config["train_micro_batch_size_per_gpu"]
        max_length = self.parallel_config.train_args.max_length

        world_size = 1 if self.parallel_config.train_args.is_partition_data else self.parallel_config.world_size
        rank = 0 if self.parallel_config.train_args.is_partition_data else self.rank

        data_engine = DataEngine(self.data_dir, tokenizer_path, micro_batch_size, max_length,world_size,rank)
        data_engine.load_data()
        return data_engine     
    
    def prepare_model(self):
        # Initialize the distributed environment.
        _init_distributed_environment(self.parallel_config, self.rank,
                                      self.distributed_init_method)
        
        # check the enabled parameter here: https://github.com/microsoft/DeepSpeed/issues/3234
        
        with deepspeed.zero.Init(config_dict_or_path=self.ds_config,
                             enabled=self.ds_config["zero_optimization"]["stage"] == 3,
                             mem_efficient_linear=False,
                             mpu=None):
            model = self.get_model()            
            model_parameters = filter(lambda p: p.requires_grad, model.parameters())
            model_engine, _, _, _ = deepspeed.initialize(model=model,
                                                         config=self.ds_config,
                                                        optimizer=None,
                                                        model_parameters=model_parameters)
            return model_engine
            
           

class DeepSpeedTrain:
    def __init__(self,parallel_config: ParallelConfig):    

        

        # get resource manually. If use num_gpus, the ray will set cuda_visible_devices automatically, and this
        # will cause the deepspeed can't get the right gpu_ids enven if we set the cuda_visible_devices manually.                
        resource_workers = [] 
        workers = []       
        
        for rank in range(parallel_config.world_size):    
            worker_cls = ResourceWorker                        
            worker_cls = ray.remote(
                        num_cpus=0,
                        num_gpus=1,
                        name=f"RW-{rank}-{parallel_config.train_args.sft_name}",
                        # resources={f"node:{master_addr}": 1e-3},
                        # runtime_env=runtime_env,                        
                    )(worker_cls).remote
            worker = worker_cls(parallel_config,rank)
            resource_workers.append(worker)

        master_addr, master_port = ray.get(resource_workers[0].get_address_and_port.remote())          
        distributed_init_method = f"tcp://{master_addr}:{master_port}"  
        print(f"deepspeed: master_addr:{master_addr},master_port:{master_port}",flush=True)

        self.resource_workers  = resource_workers        
        self.node_id_to_workers = {}
        self.node_id_to_gpus = {}
        self.node_id_to_nccl_socket_device = {}

        for resource_worker in self.resource_workers:
            device = ray.get(resource_worker.get_node_and_gpu_ids.remote())            
     
            if device.node_id not in self.node_id_to_workers:
                self.node_id_to_workers[device.node_id] = []
            
            if device.node_id not in self.node_id_to_gpus:
                self.node_id_to_gpus[device.node_id] = []    
            
            self.node_id_to_workers[device.node_id].append(resource_worker)    
            self.node_id_to_gpus[device.node_id].extend(device.gpu_ids)
            self.node_id_to_gpus[device.node_id].sort()

            if parallel_config.setup_nccl_socket_ifname_by_ip:
                self.node_id_to_nccl_socket_device[device.node_id] = ray.get(resource_worker.get_network_interface.remote())

        for node_id, resource_workers in self.node_id_to_workers.items():
            for local_rank,resource_worker in enumerate(resource_workers):
                rank = ray.get(resource_worker.rank.remote()) 
                worker_cls = Worker  
                gpu_ids = self.node_id_to_gpus[node_id]
                nccl_socket_ifname = DEFAULT_NCCL_SOCKET_IFNAME
                if parallel_config.setup_nccl_socket_ifname_by_ip:
                    nccl_socket_ifname = self.node_id_to_nccl_socket_device[node_id]
                env_vars = {
                        "RAY_EXPERIMENTAL_NOSET_CUDA_VISIBLE_DEVICES":"true",
                        "CUDA_VISIBLE_DEVICES": ",".join([str(gid) for gid in gpu_ids]),
                        "LOCAL_RANK": str(local_rank),
                        "RANK": str(rank),
                        "NCCL_SOCKET_IFNAME": f"={nccl_socket_ifname}",
                        "LOCAL_WORLD_SIZE": str(len(gpu_ids)),
                        "WORLD_SIZE": str(parallel_config.world_size)
                        }                                        
                runtime_env = {"env_vars": env_vars}  
                node_ip = ray.get(resource_worker.get_node_ip_address.remote())  
                worker_cls = ray.remote(
                            num_cpus=0, 
                            num_gpus=0,
                            name=f"W-{rank}-{parallel_config.train_args.sft_name}",
                            resources={f"node:{node_ip}": 1e-3},
                            runtime_env=runtime_env,                        
                        )(worker_cls).remote
                worker = worker_cls(parallel_config,rank,distributed_init_method)
                workers.append(worker)  
                if rank == 0:
                    addr_port = ray.get(worker.setup_tensorboard.remote())
                    if addr_port:
                        print(f"tensorboard: http://{addr_port[0]}:{addr_port[1]}",flush=True)

        self.workers = workers                              
    
              
    def _run_workers(
        self,
        method: str,
        *args,
        get_all_outputs: bool = False,
        **kwargs,
    ) -> Any:
        """Runs the given method on all workers."""
        all_outputs = []
        for worker in self.workers:
            executor = getattr(worker, method)
            executor = executor.remote
            output = executor(*args, **kwargs)
            all_outputs.append(output)

        all_outputs = ray.get(all_outputs)            
        
        if get_all_outputs:
            return all_outputs

        # Make sure all workers have the same results.
        output = all_outputs[0]
        for other_output in all_outputs[1:]:
            assert output == other_output
        return output 

class DeepSpeedTrainer:
    def __init__(self,name:str) -> None:
        self.sft_name = name
        self.dst = None 
        self.output_dir = None

    def get_checkpoint_path(self):
        return self.output_dir    

    def sfft_train(self,data_refs:List[DataServer],train_params:Dict[str,str],sys_conf: Dict[str, str]):                

        localPathPrefix = train_params.get("localPathPrefix","/tmp/byzerllm")

        sft_name = self.sft_name
        rd = f"{sft_name}-{str(uuid.uuid4())}"        

        num_gpus = int(sys_conf.get("num_gpus",0))
        
        assert num_gpus > 0, 'num_gpus must be greater than 0. Try to fix it with `!byzerllm setup "num_gpus=4"`'
        
        is_partition_data = len(data_refs) != 0

        if is_partition_data:
            assert num_gpus == len(data_refs), f'''The number of data refs({len(data_refs)}) must be equal to the number of GPUs({num_gpus}).
            Try to fix it with `!byzerllm setup "num_gpus={len(data_refs)}"` or repartition the data with the following command:
            
            ```
            run oldTable as TableRepartition.`` where partitionNum="{num_gpus}" as newTable;
            ```
            Notice that make sure Byzer engine have CPUs more than {num_gpus}.
            '''

        data_dir = train_params["localDataDir"] if "localDataDir" in train_params else os.path.join(localPathPrefix,rd,"finetune_data")
        output_dir = os.path.join(localPathPrefix,rd,"finetune_model")
        self.output_dir = output_dir
        tensorboard_dir = os.path.join(localPathPrefix,rd,"tensorboard_dir")
        model_dir = os.path.join(localPathPrefix,rd,"pretrained_model")
        
        if "localModelDir" in train_params:
            model_dir = train_params["localModelDir"]

        pretrained_model_type = train_params.get("pretrainedModelType","")
        if "/" in  pretrained_model_type:
            pretrained_model_type = pretrained_model_type.split("/")[-1]
        
        def get_model():
            if pretrained_model_type == "llama2":            
                return AutoModelForCausalLM.from_pretrained(model_dir,
                                                            trust_remote_code=True,
                                                            ignore_mismatched_sizes=True)
            else:
                return AutoModelForCausalLM.from_pretrained(model_dir,trust_remote_code=True)
        
        setup_nccl_socket_ifname_by_ip = False
        if "sfft.bool.setup_nccl_socket_ifname_by_ip" in train_params:
            setup_nccl_socket_ifname_by_ip = train_params["sfft.bool.setup_nccl_socket_ifname_by_ip"] == "true"
        
        tokenizer_path = train_params["sfft.str.tokenizer_path"] if "sfft.str.tokenizer_path" in train_params else f"{model_dir}/tokenizer.model"        
        max_length = int(train_params.get("sfft.int.max_length",4096))
        epoches = int(train_params.get("sfft.int.epoches",1))
        steps_per_epoch = int(train_params.get("sfft.int.steps_per_epoch",10))
        
        try:
            ds_config=  json.loads(train_params.get("deepspeedConfig",DEFUALT_CONFIG))
        except Exception as e:        
            print(f'deepspeedConfig is not a valid json string:\n{train_params.get("deepspeedConfig","{}")}',flush=True)
            print(f"Byzer-LLM will ues the default deepspeed config:\n{DEFUALT_CONFIG}",flush=True)
            ds_config = json.loads(DEFUALT_CONFIG)
            

        if "tensorboard"  in ds_config and  ds_config["tensorboard"].get("enabled",False):
            if "output_path" not in ds_config["tensorboard"]:
                ds_config["tensorboard"]["output_path"] = tensorboard_dir
                ds_config["tensorboard"]["job_name"] = sft_name

        print(f'''
    Train Configuration:
        pretrained_model_type:{pretrained_model_type} 
        model_dir:{model_dir} 
        output_dir:{output_dir}
        data_dir:{data_dir}
        is_partition_data:{is_partition_data}
        max_length:{max_length}
        epoches:{epoches}
        steps_per_epoch:{steps_per_epoch} 
        setup_nccl_socket_ifname_by_ip:{setup_nccl_socket_ifname_by_ip}   
        num_gpus:{num_gpus}            
            ''',flush=True)   
        

        dst = DeepSpeedTrain(ParallelConfig(
        data_refs = data_refs,
        num_workers = num_gpus,
        get_model = get_model,
        ds_config = ds_config,     
        setup_nccl_socket_ifname_by_ip = setup_nccl_socket_ifname_by_ip,
        train_args=TrainArgs(
            model_path=model_dir,
            tokenizer_path = tokenizer_path,
            data_dir = data_dir,  
            checkpoint_saving_path = output_dir,   
            steps_per_epoch = steps_per_epoch,
            max_length = max_length,
            epoches=epoches,
            is_partition_data = is_partition_data,
            sft_name=sft_name
            )
        ))

        self.dst = dst
        ray.get([worker.train.remote() for worker in dst.workers])
        if train_params.get("detached","true") == "true":
            return [],0
        chunks,count = ray.get(dst.workers[0].get_checkpoint.remote())
        return chunks,count


def sfft_train(data_refs:List[DataServer],train_params:Dict[str,str],sys_conf: Dict[str, str])->Generator[BlockRow,Any,Any]:
    
    current_time = datetime.datetime.now()
    formatted_time = current_time.strftime("%Y%m%d-%H-%M-%S")
    sft_name = train_params["name"] if "name" in train_params else f"sft-{sys_conf['OWNER']}-{formatted_time}"        

    detached = train_params.get("detached","true") == "true"
    options = {"name":sft_name}
    if detached:        
        options["lifetime"] = "detached"
                
    worker_cls = ray.remote(**options)(DeepSpeedTrainer).remote
    trainer = worker_cls(name=sft_name)

    if detached:
        print_flush(f"[{sft_name}] Detached mode is enabled. ")
        trainer.sfft_train.remote(data_refs,train_params,sys_conf)        
        return []
        
    chunks,obj_count = ray.get(trainer.sfft_train.remote(data_refs,train_params,sys_conf))    
    checkpoint_path = ray.get(trainer.get_checkpoint_path.remote())
    node = ray.get(trainer.dst.resource_workers[0].get_node_ip_address.remote())
    print_flush(f"The model is finised training, Please check the path: {node}:{checkpoint_path}")
    
    if obj_count == 0:    
        return []

    print_flush(f"[{sft_name}] Transform Model from Ray object store to new storage(delta lake), total refs: {obj_count}. ")
    count = 0
    for item in chunks:
        if count % 1000 == 0:
            print_flush(f"[{sft_name}] Process: {float(count)/obj_count*100}%")
        count += 1
        yield ray.get(item)




from .common import (
    load_pretrained,
    prepare_args,
    prepare_data,
    preprocess_data
)

from .seq2seq import (
    Seq2SeqDataCollatorForChatGLM,
    ComputeMetrics,
    Seq2SeqTrainerForChatGLM
)

from .config import ModelArguments

from .other import plot_loss



import json
import os
from typing import List, Dict, Any


def train_pt(
    name: str,
    memories: List[str],
    options: Dict[str, Any],
    dataset_dir: str,
    loras_dir: str,
    target_model_path: str,
    data_model_name: str,
    model_name: str,
):
    data = []
    min_samples = options.pop("min_samples", 1000)

    for memory in memories:
        item = {
            "text": memory,
        }
        data.append(item)

    if len(data) < min_samples:
        data = data * (min_samples // len(data) + 1)

    with open(f"{dataset_dir}/data.json", "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, ensure_ascii=False)

    with open(f"{dataset_dir}/dataset_info.json", "w", encoding="utf-8") as f:
        r = {"data": {"file_name": "data.json", "columns": {"prompt": "text"}}}
        f.write(json.dumps(r, indent=2, ensure_ascii=False))

    args = dict(
        stage="pt",
        do_train=True,
        model_name_or_path=target_model_path,
        dataset="data",
        dataset_dir=dataset_dir,
        cutoff_len=1024,
        max_samples=1000000,
        overwrite_cache=True,
        preprocessing_num_workers=1,
        template="llama3",
        finetuning_type="lora",
        lora_target="all",
        output_dir=f"{loras_dir}/{name}",
        overwrite_output_dir=True,
        per_device_train_batch_size=2,
        gradient_accumulation_steps=4,
        lr_scheduler_type="cosine",
        logging_steps=10,
        warmup_ratio=0.1,
        save_steps=1000,
        plot_loss=True,
        learning_rate=5e-5,
        num_train_epochs=1000.0,
        max_grad_norm=1.0,
        quantization_bit=4,
        loraplus_lr_ratio=16.0,
        fp16=True,
        ddp_timeout=180000000,
    )
    os.environ["WANDB_DISABLED"] = "true"
    from llamafactory.train import tuner

    tuner.run_exp({**args, **options})



from byzerllm.utils.client import ByzerLLM,InferBackend,Templates
import ray
import pytest

model_path = "/home/winubuntu/models/Qwen-1_8B-Chat"

class TestByzerLLMVLLMDeploy(object):
    llm = None

    def setup_class(self):
        if ray.is_initialized():
            ray.shutdown()
        
        ray.init(address="auto",namespace="default")
        
        self.llm = ByzerLLM(verbose=True)

        if self.llm.is_model_exist("chat"):
            self.llm.undeploy("chat")

        self.llm.setup_gpus_per_worker(1).setup_num_workers(1).setup_infer_backend(InferBackend.VLLM)
        self.llm.deploy(
            model_path=model_path,
            pretrained_model_type="custom/auto",
            udf_name="chat",
            infer_params={"backend.max_num_batched_tokens":8192,"backend.max_model_len":8192}
        )
        self.llm.setup_default_model_name("chat")
        self.llm.setup_template("chat",Templates.qwen())
        
        
    def teardown_class(self):
        if self.llm.is_model_exist("chat"):
            self.llm.undeploy("chat")
    
    def test_chat_oai(self):
        t = self.llm.chat_oai([{
            "content":"你好",
            "role":"user"
        }])
        assert len(t[0].output) > 1





    



# Copyright 2023 Databricks, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import click
import numpy as np
from datasets import Dataset, load_dataset
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    DataCollatorForLanguageModeling,
    PreTrainedTokenizer,
    Trainer,
    TrainingArguments,
    set_seed,
)

from .consts import (
    DEFAULT_INPUT_MODEL,
    DEFAULT_SEED,
    PROMPT_WITH_INPUT_FORMAT,
    PROMPT_NO_INPUT_FORMAT,
    END_KEY,
    INSTRUCTION_KEY,
    RESPONSE_KEY_NL,
)

logger = logging.getLogger(__name__)
ROOT_PATH = Path(__file__).parent.parent


class DataCollatorForCompletionOnlyLM(DataCollatorForLanguageModeling):
    def torch_call(self, examples: List[Union[List[int], Any, Dict[str, Any]]]) -> Dict[str, Any]:
        batch = super().torch_call(examples)

        # The prompt ends with the response key plus a newline.  We encode this and then try to find it in the
        # sequence of tokens.  This should just be a single token.
        response_token_ids = self.tokenizer.encode(RESPONSE_KEY_NL)

        labels = batch["labels"].clone()

        for i in range(len(examples)):

            response_token_ids_start_idx = None
            for idx in np.where(batch["labels"][i] == response_token_ids[0])[0]:
                response_token_ids_start_idx = idx
                break

            if response_token_ids_start_idx is None:
                raise RuntimeError(
                    f'Could not find response key {response_token_ids} in token IDs {batch["labels"][i]}'
                )

            response_token_ids_end_idx = response_token_ids_start_idx + 1

            # Make pytorch loss function ignore all tokens up through the end of the response key
            labels[i, :response_token_ids_end_idx] = -100

        batch["labels"] = labels

        return batch


def preprocess_batch(batch: Dict[str, List], tokenizer: AutoTokenizer, max_length: int) -> dict:
    return tokenizer(
        batch["text"],
        max_length=max_length,
        truncation=True,
    )


def load_training_dataset(path_or_dataset: str = "databricks/databricks-dolly-15k") -> Dataset:
    logger.info(f"Loading dataset from {path_or_dataset}")
    dataset = load_dataset(path_or_dataset)["train"]
    logger.info("Found %d rows", dataset.num_rows)

    def _add_text(rec):
        instruction = rec["instruction"]
        response = rec["response"]
        context = rec.get("context")

        if not instruction:
            raise ValueError(f"Expected an instruction in: {rec}")

        if not response:
            raise ValueError(f"Expected a response in: {rec}")

        # For some instructions there is an input that goes along with the instruction, providing context for the
        # instruction.  For example, the input might be a passage from Wikipedia and the instruction says to extract
        # some piece of information from it.  The response is that information to extract.  In other cases there is
        # no input.  For example, the instruction might be open QA such as asking what year some historic figure was
        # born.
        if context:
            rec["text"] = PROMPT_WITH_INPUT_FORMAT.format(instruction=instruction, response=response, input=context)
        else:
            rec["text"] = PROMPT_NO_INPUT_FORMAT.format(instruction=instruction, response=response)
        return rec

    dataset = dataset.map(_add_text)

    return dataset


def load_tokenizer(pretrained_model_name_or_path: str = DEFAULT_INPUT_MODEL) -> PreTrainedTokenizer:
    logger.info(f"Loading tokenizer for {pretrained_model_name_or_path}")
    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.add_special_tokens({"additional_special_tokens": [END_KEY, INSTRUCTION_KEY, RESPONSE_KEY_NL]})
    return tokenizer


def load_model(
    pretrained_model_name_or_path: str = DEFAULT_INPUT_MODEL, *, gradient_checkpointing: bool = False
) -> AutoModelForCausalLM:
    logger.info(f"Loading model for {pretrained_model_name_or_path}")
    model = AutoModelForCausalLM.from_pretrained(
        pretrained_model_name_or_path, trust_remote_code=True, use_cache=False if gradient_checkpointing else True
    )
    return model


def get_model_tokenizer(
    pretrained_model_name_or_path: str = DEFAULT_INPUT_MODEL, *, gradient_checkpointing: bool = False
) -> Tuple[AutoModelForCausalLM, PreTrainedTokenizer]:
    tokenizer = load_tokenizer(pretrained_model_name_or_path)
    model = load_model(pretrained_model_name_or_path, gradient_checkpointing=gradient_checkpointing)
    model.resize_token_embeddings(len(tokenizer))

    return model, tokenizer


def preprocess_dataset(tokenizer: AutoTokenizer, max_length: int, seed=DEFAULT_SEED) -> Dataset:
    """Loads the training dataset and tokenizes it so it is ready for training.

    Args:
        tokenizer (AutoTokenizer): Tokenizer tied to the model.
        max_length (int): Maximum number of tokens to emit from tokenizer.

    Returns:
        Dataset: HuggingFace dataset
    """

    dataset = load_training_dataset()

    logger.info("Preprocessing dataset")
    _preprocessing_function = partial(preprocess_batch, max_length=max_length, tokenizer=tokenizer)
    dataset = dataset.map(
        _preprocessing_function,
        batched=True,
        remove_columns=["instruction", "context", "response", "text", "category"],
    )

    # Make sure we don't have any truncated records, as this would mean the end keyword is missing.
    logger.info("Processed dataset has %d rows", dataset.num_rows)
    dataset = dataset.filter(lambda rec: len(rec["input_ids"]) < max_length)
    logger.info("Processed dataset has %d rows after filtering for truncated records", dataset.num_rows)

    logger.info("Shuffling dataset")
    dataset = dataset.shuffle(seed=seed)

    logger.info("Done preprocessing")

    return dataset


def train(
    *,
    input_model: str,
    local_output_dir: str,
    dbfs_output_dir: str,
    epochs: int,
    per_device_train_batch_size: int,
    per_device_eval_batch_size: int,
    lr: float,
    seed: int,
    deepspeed: str,
    gradient_checkpointing: bool,
    local_rank: str,
    bf16: bool,
    logging_steps: int,
    save_steps: int,
    eval_steps: int,
    test_size: Union[float, int],
    save_total_limit: int,
    warmup_steps: int,
):
    set_seed(seed)

    model, tokenizer = get_model_tokenizer(
        pretrained_model_name_or_path=input_model, gradient_checkpointing=gradient_checkpointing
    )

    # Use the same max length that the model supports.  Fall back to 1024 if the setting can't be found.
    # The configuraton for the length can be stored under different names depending on the model.  Here we attempt
    # a few possible names we've encountered.
    conf = model.config
    max_length = None
    for length_setting in ["n_positions", "max_position_embeddings", "seq_length"]:
        max_length = getattr(model.config, length_setting, None)
        if max_length:
            logger.info(f"Found max lenth: {max_length}")
            break
    if not max_length:
        max_length = 1024
        logger.info(f"Using default max length: {max_length}")

    processed_dataset = preprocess_dataset(tokenizer=tokenizer, max_length=max_length, seed=seed)

    split_dataset = processed_dataset.train_test_split(test_size=test_size, seed=seed)

    logger.info("Train data size: %d", split_dataset["train"].num_rows)
    logger.info("Test data size: %d", split_dataset["test"].num_rows)

    data_collator = DataCollatorForCompletionOnlyLM(
        tokenizer=tokenizer, mlm=False, return_tensors="pt", pad_to_multiple_of=8
    )

    if not dbfs_output_dir:
        logger.warn("Will NOT save to DBFS")

    training_args = TrainingArguments(
        output_dir=local_output_dir,
        per_device_train_batch_size=per_device_train_batch_size,
        per_device_eval_batch_size=per_device_eval_batch_size,
        fp16=False,
        bf16=bf16,
        learning_rate=lr,
        num_train_epochs=epochs,
        deepspeed=deepspeed,
        gradient_checkpointing=gradient_checkpointing,
        logging_dir=f"{local_output_dir}/runs",
        logging_strategy="steps",
        logging_steps=logging_steps,
        evaluation_strategy="steps",
        eval_steps=eval_steps,
        save_strategy="steps",
        save_steps=save_steps,
        save_total_limit=save_total_limit,
        load_best_model_at_end=False,
        report_to="tensorboard",
        disable_tqdm=True,
        remove_unused_columns=False,
        local_rank=local_rank,
        warmup_steps=warmup_steps,
    )

    logger.info("Instantiating Trainer")

    trainer = Trainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        train_dataset=split_dataset["train"],
        eval_dataset=split_dataset["test"],
        data_collator=data_collator,
    )

    logger.info("Training")
    trainer.train()

    logger.info(f"Saving Model to {local_output_dir}")
    trainer.save_model(output_dir=local_output_dir)

    if dbfs_output_dir:
        logger.info(f"Saving Model to {dbfs_output_dir}")
        trainer.save_model(output_dir=dbfs_output_dir)

    logger.info("Done.")


@click.command()
@click.option("--input-model", type=str, help="Input model to fine tune", default=DEFAULT_INPUT_MODEL)
@click.option("--local-output-dir", type=str, help="Write directly to this local path", required=True)
@click.option("--dbfs-output-dir", type=str, help="Sync data to this path on DBFS")
@click.option("--epochs", type=int, default=3, help="Number of epochs to train for.")
@click.option("--per-device-train-batch-size", type=int, default=8, help="Batch size to use for training.")
@click.option("--per-device-eval-batch-size", type=int, default=8, help="Batch size to use for evaluation.")
@click.option(
    "--test-size", type=int, default=1000, help="Number of test records for evaluation, or ratio of test records."
)
@click.option("--warmup-steps", type=int, default=None, help="Number of steps to warm up to learning rate")
@click.option("--logging-steps", type=int, default=10, help="How often to log")
@click.option("--eval-steps", type=int, default=50, help="How often to run evaluation on test records")
@click.option("--save-steps", type=int, default=400, help="How often to checkpoint the model")
@click.option("--save-total-limit", type=int, default=10, help="Maximum number of checkpoints to keep on disk")
@click.option("--lr", type=float, default=1e-5, help="Learning rate to use for training.")
@click.option("--seed", type=int, default=DEFAULT_SEED, help="Seed to use for training.")
@click.option("--deepspeed", type=str, default=None, help="Path to deepspeed config file.")
@click.option(
    "--gradient-checkpointing/--no-gradient-checkpointing",
    is_flag=True,
    default=True,
    help="Use gradient checkpointing?",
)
@click.option(
    "--local_rank",
    type=str,
    default=True,
    help="Provided by deepspeed to identify which instance this process is when performing multi-GPU training.",
)
@click.option("--bf16", type=bool, default=True, help="Whether to use bf16 (preferred on A100's).")
def main(**kwargs):
    train(**kwargs)


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s %(levelname)s [%(name)s] %(message)s", level=logging.INFO, datefmt="%Y-%m-%d %H:%M:%S"
    )
    try:
        main()
    except Exception:
        logger.exception("main failed")
        raise
